from typing import Literal, TypeAlias, Dict, Callable, List, Tuple, Any


ToolsDefType: TypeAlias = Dict[Callable, List[Tuple[str, Dict[str, Any], bool]]]
Role: TypeAlias = Literal['user', 'assistant', 'tool', 'system']
Framework: TypeAlias = Literal['openai', 'anthropic', 'mistral']

Color: TypeAlias = Literal[
    "red",
    "green",
    "yellow",
    "blue",
    "magenta",
    "cyan"
]



import re
import os
import inspect
import requests
from tqdm import tqdm
from bs4 import BeautifulSoup
from duckduckgo_search import DDGS
from googleapiclient.discovery import build
from typing import List, Tuple
from .color_logger import get_logger
from .settings import Settings
from ._types import ToolsDefType


_GOOGLE_SEARCH_API_KEY = os.environ.get('GOOGLE_SEARCH_API_KEY', None)
_GOOGLE_SEARCH_CSE_ID = os.environ.get('GOOGLE_SEARCH_CSE_ID', None)


def handle_tool_error(e: Exception) -> None:
    get_logger().error(f'ERROR [{e.__class__.__name__} in {inspect.stack()[1].function}]: {str(e)}', color='red')


def web_search(query: str, max_results: int = 10) -> str:
    """
    Search the web for the provided query, and returns the title, URL and description of the results.
    Search results are separated by: =====

    Example of two search results:

    Title: Welcome to My Site
    URL: http://www.mysite.xyz
    Description: This is my private website, see my stuff here
    =====
    Title: Grapes Online
    URL: http://www.grapes.com
    Description: This is the number one site for grapes fans and lovers
    """
    if max_results < 1: max_results = 1
    elif max_results > 10: max_results = 10

    def duckduckgo_search(query: str, max_results: int) -> List[Tuple[str, str, str]]:
        results = []
        ddgs = DDGS() 
        results = ddgs.text(query, max_results=max_results)
        for result in tqdm(results):
            url = result['href']
            if url is None:
                continue
            body = result['body']
            title = result['title'] or ''
            results.append((title, url, body))
        return results
    
    def google_search(query: str, max_results: int, api_key: str, cse_id: str) -> List[Tuple[str, str, str]]:
        results = []
        service = build("customsearch", "v1", developerKey=api_key)
        response = service.cse().list(q=query, cx=cse_id, num=max_results).execute()
        for result in tqdm(response['items']):
            results.append((result['title'], result['link'], result['snippet']))
        return results

    try:
        if _GOOGLE_SEARCH_API_KEY and _GOOGLE_SEARCH_CSE_ID:
            get_logger().debug('Searching the web with Google')
            results = google_search(query=query, max_results=max_results, api_key=_GOOGLE_SEARCH_API_KEY, cse_id=_GOOGLE_SEARCH_CSE_ID)
        else:
            get_logger().debug('Searching the web with DuckDuckGo')
            results = duckduckgo_search(query=query, max_results=max_results)
        
        if results:
            return '\n=====\n'.join(f"Title: {title}\nURL: {url}\nDescription: {body}\n" for (title, url, body) in results)
        else:
            return "No results"

    except Exception as e:
        handle_tool_error(e)
        return f"Error while searching the web: {e}"


def visit_website(url: str) -> str:
    """
    Goes to the provided URL and returns a simple version of the page text. Images and styling are excluded.
    """
    try:
        headers = {'User-Agent': Settings().web.user_agent}
        response = requests.get(url, timeout=Settings().web.surf_timeout_seconds, headers=headers)
        if response.status_code == 200:
            soup = BeautifulSoup(response.content, 'html.parser')
            text = soup.get_text()

            # Clean whitespaces
            text = re.sub(pattern=r'[ \t]+', repl=' ', string=text)  # Replace multiple spaces and tabs with a single space
            text = re.sub(pattern=r'\n{3,}', repl='\n\n', string=text)  # Replace more than two newlines with two newlines
            text = text.strip()

            return text 
        else:
            raise RuntimeError(f"ERROR: Failed to retrieve the webpage. Status code: {response.status_code}")
    except Exception as e:
        handle_tool_error(e)
        return f"ERROR: {e}"



tools_params_definitions: ToolsDefType = {
    web_search: [("query", {"type": "string", "description": "The query to search on the web"}, True),
                 ("max_results", {"type": "number", "description": "Maximal number of results to retrieve. Must be between 1 and 10, default is 10."}, False)],
    visit_website: [("url", {"type": "string", "description": "The URL of the page to scrape"}, True)],
}


import json
from typing import List, Dict, Any, Generator
from dataclasses import dataclass
from mistralai.models.chat_completion import ChatMessage as MistralChatMessage, ToolCall as MistralToolCall, FunctionCall as MistralFunctionCall
from ._types import Role, Framework


@dataclass
class ToolCall:
    call_id: str
    function: str
    arguments: Dict[str, Any]

    def __str__(self) -> str:
        return f"{self.function}: {json.dumps(self.arguments)}"
    
    def as_dict(self) -> Dict[str, Any]:
        return {
            'call_id': self.call_id,
            'function': self.function,
            'arguments': self.arguments
        }


class Message:
    def __init__(self, role: Role, content: str | None = None, tool_calls: List[ToolCall] = [], *, model: str | None = None) -> None:
        if role == 'assistant' and not model:
            raise ValueError('AI generated messages must be provided with model name')
        self.role: Role = role
        self.content: str | None = content
        self.tool_calls: List[ToolCall] = tool_calls
        self.model: str | None = model
    
    def __str__(self) -> str:
        return f"{{ role: {self.role}, content: '{self.content or ''}', tool_calls: [{', '.join([str(t) for t in self.tool_calls])}]}}"
    
    def as_dict(self) -> Dict[str, Any]:
        dct = {
            'role': self.role,
            'content': self.content
        }
        if self.tool_calls:
            dct['tool_calls'] = [t.as_dict() for t in self.tool_calls]
        if self.model:
            dct['model'] = self.model
        return dct
    
    @property
    def displayed(self) -> bool:
        if self.role == 'user' or (self.role == 'assistant' and not self.tool_calls):
            return True
        else:
            return False
    
    def update(self, content: str | None = None, tool_calls: List[ToolCall] | None = None) -> None:
        if content:
            self.content = (self.content or '') + content
        if tool_calls:
            self.tool_calls = (self.tool_calls or []) + tool_calls

    def edit(self, content: str | None = None, tool_calls: List[ToolCall] | None = None) -> None:
        if content:
            self.content = content
        if tool_calls:
            self.tool_calls = tool_calls
    
    def to(self, framework: Framework):
        match framework:
            case 'openai':
                if self.role == 'tool':
                    tool_call_id = self.tool_calls[0].call_id
                    tool_calls = None
                elif self.tool_calls:
                    tool_call_id = None
                    tool_calls = list()
                    for tool_call in self.tool_calls:
                        tool_call_dict = dict()
                        tool_call_dict['id'] = tool_call.call_id
                        tool_call_dict['type'] = 'function'
                        tool_call_dict['function'] = {'name': tool_call.function, 'arguments': str(tool_call.arguments)}
                        tool_calls.append(tool_call_dict)
                else:
                    tool_calls = None
                    tool_call_id = None
                message = {
                    'role': self.role,
                    'content': self.content,
                    }
                if tool_calls:
                    message['tool_calls'] = tool_calls
                if tool_call_id:
                    message['tool_call_id'] = tool_call_id
                return message
            
            case 'anthropic':
                if self.role == 'system':
                    return None
                message = {'role': self.role, 'content': self.content}
                if not message['content'] and self.tool_calls:
                    message['content'] = f"Executing functions: [{', '.join([str(t) for t in self.tool_calls])}]"
                elif message['role'] == 'tool':
                    message['role'] = 'user'
                return message

            case 'mistral':
                if self.role in ['system', 'user']:
                    message = MistralChatMessage(role=self.role, content=self.content or '')
                elif self.role == 'tool':
                    message = MistralChatMessage(role=self.role, content=self.content or '', name=self.tool_calls[0].function)
                elif self.role == 'assistant':
                    if self.tool_calls:
                        mistral_tool_calls = []
                        for tool_call in self.tool_calls:
                            function_call = MistralFunctionCall(name=tool_call.function, arguments=json.dumps(tool_call.arguments))
                            mistral_tool_calls.append(MistralToolCall(function=function_call))
                    else:
                        mistral_tool_calls = None
                    message = MistralChatMessage(role=self.role, content=self.content or '', tool_calls=mistral_tool_calls)
                return message


class Messages(list[Message]):
    def __init__(self, *args) -> None:
        super().__init__(*args)

    @classmethod
    def from_dict(cls, messages_as_dict: List[Dict[str, Any]]):
        messages = cls()
        for message in messages_as_dict:
            tool_calls: List[ToolCall] = list()
            if 'tool_calls' in message:
                for tool_call in message['tool_calls']:
                    tool_calls.append(
                        ToolCall(
                            call_id=tool_call['call_id'],
                            function=tool_call['function'],
                            arguments=tool_call['arguments']
                        )
                    )
            messages.append(
                role=message['role'],
                content=message['content'],
                tool_calls=tool_calls,
                model=message.get('model', None)
            )
        return messages

    @property
    def empty(self) -> bool:
        return not bool(self)
    
    @property
    def last_message_index(self) -> int:
        if self.empty:
            raise ValueError('No messages!')
        return len(self)-1
    
    @property
    def system_prompt(self) -> str | None:
        if self.empty:
            return None
        message = self[0]
        if message.role == 'system':
            return message.content
        else:
            return None
    
    def as_dict(self) -> List[Dict[str, Any]]:
        return [m.as_dict() for m in self]

    def append(self, role: Role, content: str | None, tool_calls: List[ToolCall] = [], model: str | None = None) -> None:
        super().append(Message(role, content, tool_calls=tool_calls, model=model))
    
    def to(self, framework: Framework):
        msgs = [m.to(framework) for m in self]
        return [m for m in msgs if m]
    
    def update(self, i: int, /, content: str | None = None, tool_calls: List[ToolCall] | None = None) -> None:
        self[i].update(content, tool_calls)

    def edit(self, i: int, /, content: str | None = None, tool_calls: List[ToolCall] | None = None) -> None:
        self[i].edit(content, tool_calls)  



import os
import json
import pathlib
import gradio as gr
from datetime import datetime
from typing import List, Tuple, Generator, Set, Tuple, Dict, Any
from .saved_chat import SavedChat
from .chatbot import Chatbot
from .settings import Settings
from .utils import path_to_resource
from ._types import Framework


class UI:
    CHAT_FILE_TIME_FORMAT = "%d/%m/%Y, %H:%M:%S"
    MODEL_NAME_SEPARATOR = "\n\n🤖 "

    def __init__(self, chatbot: Chatbot, available_frameworks: Set[Framework], port: int) -> None:
        self.ui: gr.Blocks | None = None
        self.chatbot = chatbot
        self.available_frameworks = available_frameworks
        self._generate_text = True
        self.preferences: Dict[str, Any] = dict()
        self.port: int = port

    def __enter__(self) -> gr.Blocks:
        self.preferences = self._load_preferences_from_file()
        self.ui = self._build_ui()
        self.ui.launch(favicon_path=path_to_resource("eevee_50.png"), inbrowser=True, show_error=True, server_port=self.port)
        return self.ui

    def __exit__(self, *args) -> None:
        if self.ui:
            self.ui.close()
            with open(self.preferences_file_path, 'w') as f:
                json.dump(self.preferences, f)

    @property
    def preferences_file_path(self) -> str:
        filename = "pref.json"
        current_dir = pathlib.Path(__file__).parent.resolve()
        return os.path.join(current_dir, filename)

    def _load_preferences_from_file(self) -> Dict[str, Any]:
        try:
            with open(self.preferences_file_path, 'r') as f:
                return json.load(f)
        except:
            return {}

    def _get_list_of_models(self) -> List[Tuple[Framework, str]]:
        available_models: List[Tuple[Framework, str]] = list()
        for framework in self.available_frameworks:
            models: List[str] = Settings().models[framework]
            available_models += [(framework, model) for model in models]
        return available_models
    
    def _stop_text_generation(self) -> None:
        self._generate_text = False

    def _add_user_message_to_chat(self, prompt: str, history: List[List[str | None]]) -> Tuple[str, List[List[str | None]]]:
        return '', history + [[prompt, None]]

    def _add_bot_message_to_chat(self, history: List[List[str | None]], model: str, temperature: float, as_json: bool, system_prompt: str) -> Generator:
        self._generate_text = True
        if as_json:
            generator = self.chatbot.get_json_response(history[-1][0] or '', system_prompt=system_prompt, model=model, temperature=temperature)
        else:
            generator = self.chatbot.get_stream_response(history[-1][0] or '', system_prompt=system_prompt, model=model, temperature=temperature)
        
        for chunk in generator:
            if not self._generate_text:
                break
            if chunk.startswith(self.chatbot.INFO_TOKEN):
                gr.Info(chunk.strip(self.chatbot.INFO_TOKEN))
            elif chunk.startswith(self.chatbot.WARNING_TOKEN):
                gr.Warning(chunk.strip(self.chatbot.WARNING_TOKEN))
            else:
                chunk, model = chunk.split(self.chatbot.METADATA_TOKEN, 1)
                current_message: str = history[-1][1] or ''
                current_message = self.MODEL_NAME_SEPARATOR.join(current_message.split(self.MODEL_NAME_SEPARATOR)[:-1])
                current_message += chunk
                current_message += f'{self.MODEL_NAME_SEPARATOR}_{model}_'
                history[-1][1] = current_message
                yield history 

    def _undo_last_message(self, history: List[List[str]]) -> List[List[str]]:
        self._generate_text = False
        history.pop()
        self.chatbot.delete_last_interaction()
        return history

    def _start_new_chat(self) -> Tuple[str, List]:
        self._generate_text = False
        self.chatbot.reset_chat()
        return '', []

    def _title_and_time_to_chat_display_name(self, title: str, time: datetime) -> str:
        return f'{title} ({time.strftime(self.CHAT_FILE_TIME_FORMAT)})'

    def _display_name_to_title_and_time(self, display_name: str) -> Tuple[str, datetime]:
        pieces = display_name.split(' (')
        title = ' ('.join(pieces[:-1])
        time = datetime.strptime(pieces[-1][:-1], self.CHAT_FILE_TIME_FORMAT)
        return title, time

    def _save_chat(self) -> None:
        self.chatbot.export_chat()

    def _load_chat(self, display_name: str) -> Tuple[None, List[List[str]]]:
        history: List[List[str]] = list()
        title, start_time = self._display_name_to_title_and_time(display_name)
        self.chatbot.load_chat(title, start_time)
        messages = self.chatbot.get_displayed_messages()
        for message in messages:
            if message.role == 'user':
                history.append([message.content or '', ''])
            elif message.role == 'assistant':
                if message.content:
                    history[-1][1] = message.content + (f'{self.MODEL_NAME_SEPARATOR}_{message.model}_' if message.model else '')
                else:
                    history[-1][1] = ''
            else:
                raise ValueError(f"Can't display message with role {message.role}")
        return None, history
    
    def _delete_chat_file(self, display_name: str) -> None:
        title, start_time = self._display_name_to_title_and_time(display_name)
        filename = SavedChat.title_and_time_to_filename(title, start_time)
        file_path = os.path.join(self.chatbot.saved_chats_dir, filename)
        pathlib.Path.unlink(file_path)  # type: ignore

    def _list_saved_chats(self) -> List[str]:
        saved_chats = self.chatbot.list_saved_chats()
        return [self._title_and_time_to_chat_display_name(title, time) for (title, time) in saved_chats]
    
    def _display_name_of_framework(self, framework: Framework) -> str:
        match framework:
            case 'openai':
                return 'OpenAI'
            case 'deepseek':
                return 'DeepSeek'
            case _:
                return framework.capitalize()

    def _build_ui(self) -> gr.Blocks:
        scrollable_checkbox_group_css = """
            .files_list {
                height: 30vh;
                min-height: 150px;
                width: 100%;
                overflow-x: auto !important;
                overflow-y: auto !important;
                scrollbar-width: thin !important;
            }  
        """
        available_models = self._get_list_of_models()
        available_models = [(f'{self._display_name_of_framework(framework)}: {model}', model) for (framework, model) in available_models]
        available_models = sorted(available_models, key=lambda t: t[0])
        preferred_model = self.preferences.get('model', '')
        if preferred_model not in [t[1] for t in available_models]:
            self.preferences['model'] = available_models[0][1]

        with gr.Blocks(title="Eevee Chat", css=scrollable_checkbox_group_css, theme=gr.themes.Base()) as ui:
            with gr.Row():
                with gr.Column(scale=10):
                    gr.Markdown(f"# 💬 Eevee Chat")
                with gr.Column(scale=1):
                    new_chat = gr.Button("New Chat")
            
            with gr.Row():
                with gr.Column(scale=1, variant='panel'):
                    with gr.Group():
                        model = gr.Dropdown(label="Model", interactive=True, choices=available_models, value=self.preferences['model'])  # type: ignore
                        with gr.Accordion(label="System Prompt", open=False):
                            system_prompt = gr.TextArea(value="You are a helpful AI assistance, and your task is to assist the user with all its requests in the best possible way", container=False, interactive=True, lines=10)
                        temperature = gr.Slider(label="Temperature", minimum=0., maximum=1., step=.01, value=self.preferences.get('temperature', 0.))
                        force_json = gr.Checkbox(label="Force JSON", value=False, interactive=True)
                    gr.Markdown("Not all models support all options, see documentation for more information")
                    gr.Markdown("\n---\n")
                    with gr.Group():
                        saved_chats = gr.Radio(label="Saved Chats", choices=self._list_saved_chats(), elem_classes="files_list", value=None)  # type: ignore
                        load_chat = gr.Button("Load")
                        delete_chat = gr.Button("Delete", variant='stop')

                with gr.Column(scale=10):
                    chat = gr.Chatbot(show_label=False, show_copy_button=True, height='80vh')
                    with gr.Row():
                        undo_last = gr.Button("Undo Last", variant='stop')
                        msg = gr.Textbox(label="Prompt (Press Enter to Send)", show_label=False, scale=9, container=False)
                        submit = gr.Button("Submit", variant='primary', scale=1)
                        stop = gr.Button("🟥 Stop", variant='stop', scale=1, visible=False)

            submit.click(
                self._add_user_message_to_chat, [msg, chat], [msg, chat]
            ).then(
                lambda: (gr.update(visible=True), gr.update(visible=False)), None, [stop, submit]
            ).then(
                self._add_bot_message_to_chat, [chat, model, temperature, force_json, system_prompt], chat
            ).then(
                self._save_chat
            ).then(
                lambda: (gr.update(visible=True), gr.update(visible=False)), None, [submit, stop]
            ).then(
                lambda: gr.update(choices=self._list_saved_chats()), None, saved_chats
            )

            msg.submit(
                self._add_user_message_to_chat, [msg, chat], [msg, chat]
            ).then(
                lambda: (gr.update(visible=True), gr.update(visible=False)), None, [stop, submit]
            ).then(
                self._add_bot_message_to_chat, [chat, model, temperature, force_json, system_prompt], chat
            ).then(
                self._save_chat
            ).then(
                lambda: (gr.update(visible=True), gr.update(visible=False)), None, [submit, stop]
            ).then(
                lambda: gr.update(choices=self._list_saved_chats()), None, saved_chats
            )

            stop.click(self._stop_text_generation)
            undo_last.click(self._undo_last_message, chat, chat).then(self._save_chat)
            new_chat.click(self._start_new_chat, None, [msg, chat])
            load_chat.click(self._start_new_chat, None, [msg, chat]).then(self._load_chat, saved_chats, [saved_chats, chat])
            delete_chat.click(self._delete_chat_file, saved_chats, None).then(lambda: gr.update(choices=self._list_saved_chats()), None, saved_chats)

            def __update_pref_model(model: str) -> None: self.preferences['model'] = model
            def __update_pref_temperature(temperature: float) -> None: self.preferences['temperature'] = temperature
            model.change(__update_pref_model, model, None)
            temperature.change(__update_pref_temperature, temperature, None)

        return ui
    


import os
import json
from datetime import datetime
from typing import Dict, Any, Tuple
from .messages import Messages


class SavedChat:
    FILE_PREFIX = "chat_"
    FILE_SUFFIX = ".json"
    TIME_FORMAT = "%Y-%m-%d-%H-%M-%S"
    MAX_NUM_OF_WORDS_IN_TITLE = 8

    def __init__(self, 
                 messages: Messages, 
                 start_time: datetime, 
                 directory: str, 
                 title: str | None = None,
                 **metadata
                 ) -> None:
        self._messages: Messages = messages
        self._start_time: datetime = start_time
        self._title: str = title or self._create_title()
        self._path: str = os.path.join(directory, self.title_and_time_to_filename(self._title, self._start_time))
        self._metadata: Dict[str, Any] = metadata

    @classmethod
    def from_chat_file(cls, file_path: str):
        with open(file_path, 'r') as f:
            loaded_file: Dict[str, Any] = json.load(f)
        messages = Messages.from_dict(loaded_file['messages'])
        metadata = loaded_file.get('metadata', {})
        pieces = file_path.split('/')
        directory = '/'.join(pieces[:-1])
        title, start_time = cls.filename_to_title_and_time(pieces[-1])
        return cls(messages, start_time, directory, title, **metadata)
    
    @classmethod
    def from_chat_title_and_time(cls, title: str, start_time: datetime, directory: str):
        filename = cls.title_and_time_to_filename(title, start_time)
        file_path = os.path.join(directory, filename)
        with open(file_path, 'r') as f:
            loaded_file: Dict[str, Any] = json.load(f)
        messages = Messages.from_dict(loaded_file['messages'])
        metadata = loaded_file.get('metadata', {})  
        return cls(messages, start_time, directory, title, **metadata)
    
    @property
    def messages(self) -> Messages:
        return self._messages
    
    @property
    def start_time(self) -> datetime:
        return self._start_time
    
    @property
    def path(self) -> str:
        return self._path
    
    @property
    def title(self) -> str:
        return self._title
    
    @classmethod
    def title_and_time_to_filename(cls, title: str, start_time: datetime) -> str:
        return cls.FILE_PREFIX + title.replace(' ', '_') + '_' + start_time.strftime(cls.TIME_FORMAT) + cls.FILE_SUFFIX
    
    @classmethod
    def filename_to_title_and_time(cls, filename: str) -> Tuple[str, datetime]:
        stripped = filename[len(cls.FILE_PREFIX):-len(cls.FILE_SUFFIX)]
        pieces = stripped.split('_')
        time = datetime.strptime(pieces[-1], cls.TIME_FORMAT)
        title = ' '.join(pieces[:-1])
        return title, time
    
    def from_metadata(self, key: str) -> Any:
        return self._metadata.get(key, None)
    
    def _create_title(self) -> str:
        first_message_content = self.messages[1].content if self.messages.system_prompt is not None else self.messages[0].content
        if not first_message_content:
            title = 'untitled'
        else:  
            maxsplit = self.MAX_NUM_OF_WORDS_IN_TITLE + 1
            first_words = first_message_content.split(' ', maxsplit=maxsplit)
            title = ' '.join(first_words[:maxsplit])
            title = ''.join(e for e in title if (e.isalnum() or e==' '))
            if len(first_words) > maxsplit:
                title += '...'
        return title
    
    def save_to_file(self) -> None:
        file_as_dict = {
            "start_time": self.start_time.strftime(self.TIME_FORMAT),
            "messages": self.messages.as_dict(),
            "metadata": self._metadata
        }
        with open(self.path, 'w') as f:
            f.write(json.dumps(file_as_dict, indent=4))



import os
import pathlib
import traceback
from datetime import datetime
from typing import Set, Dict, List, Generator, Any, Tuple
from .messages import Messages, Message, ToolCall
from .tools import tools_params_definitions
from .color_logger import get_logger
from .framework_models import get_model_framework
from .saved_chat import SavedChat
from .chat_connectors.connector_interface import Connector
from .chat_connectors.openai_connector import OpenAIConnector
from .chat_connectors.anthropic_connector import AnthropicConnector
from .chat_connectors.mistral_connector import MistralConnector
from ._types import Framework


class Chatbot:
    METADATA_TOKEN = "$%^"

    def __init__(self, available_frameworks: Set[Framework]) -> None:
        if not available_frameworks:
            raise RuntimeError('No available frameworks found! Make sure you supplied API keys')

        self.clients: Dict[Framework, Connector] = dict()
        self.callables = {f.__name__: f for f in tools_params_definitions.keys()}
        self.tools = self._build_tools()
        self.messages = Messages()
        self.logger = get_logger()
        self.start_time = datetime.now()

        for framework in available_frameworks:
            match framework:
                case 'openai':
                    client = OpenAIConnector()
                case 'mistral':
                    client = MistralConnector()
                case 'anthropic':
                    client = AnthropicConnector()
            self.clients[framework] = client

        self.INFO_TOKEN = self.clients[list(self.clients.keys())[0]].INFO_TOKEN
        self.WARNING_TOKEN = self.clients[list(self.clients.keys())[0]].WARNING_TOKEN

    def _build_tools(self) -> List[Dict[str, Any]]:
        tools = list()
        for func, v in tools_params_definitions.items():
            params = {}
            required = []
            for p in v:
                params[p[0]] = p[1]
                if p[2]: 
                    required.append(p[0])

            tools.append({
                "type": "function",
                "function": {
                    "name": func.__name__,
                    "description": func.__doc__,
                    "parameters": {
                        "type": "object",
                        "properties": params
                    },
                    "required": required
                }
            })
        return tools 

    @property
    def saved_chats_dir(self) -> str:
        SAVED_CHATS_DIR = "saved_chats"
        current_dir = pathlib.Path(__file__).parent.resolve()
        directory = os.path.join(current_dir, SAVED_CHATS_DIR)
        if not os.path.exists(directory):
            os.makedirs(directory)
        return directory
    
    def reset_chat(self) -> None:
        self.messages = Messages()
        self.start_time = datetime.now()

    def delete_last_interaction(self) -> None:
        if self.messages.empty: return None
        self.messages.pop()
        last_message = self.messages[-1]
        while not last_message.displayed:
            if last_message.role == 'system' or self.messages.empty:
                break
            self.messages.pop()
            last_message = self.messages[-1]

    def _prepare_for_response(self, prompt: str, system_prompt: str) -> None:
        if self.messages.empty:
            self.messages.append("system", system_prompt)
        else:
            self.messages.edit(0, content=system_prompt)
        self.messages.append('user', prompt)      

    def get_stream_response(self, prompt: str, *, system_prompt: str, model: str, temperature: float) -> Generator[str, None, None]:        
        framework = get_model_framework(model)
        self._prepare_for_response(prompt=prompt, system_prompt=system_prompt)

        try:
            final_message = False
            while not final_message:
                final_message = True
                generator = self.clients[framework].get_streaming_response(model=model, temperature=temperature, messages=self.messages, tools=self.tools)
                for token in generator:
                    if isinstance(token, list):
                        # Only option is list of ToolCall
                        tool_calls: List[ToolCall] = token
                        final_message = False
                        if self.messages[self.messages.last_message_index].role == 'assistant':
                            self.messages.update(self.messages.last_message_index, tool_calls=tool_calls)
                        else:
                            self.messages.append('assistant', content=None, tool_calls=tool_calls, model=model)
                        for tool_call in tool_calls:
                            yield self.INFO_TOKEN + "Running tool: " + tool_call.function
                            self.logger.info(f"Running tool {tool_call.function}: {str(tool_call.arguments)}", color='yellow')
                            tool_output = self.callables[tool_call.function](**tool_call.arguments)
                            self.logger.debug("Tool output:\n" + tool_output)
                            self.messages.append(role='tool', content=tool_output, tool_calls=[tool_call])
                    else:
                        if self.messages[self.messages.last_message_index].role == 'assistant':
                            self.messages.update(self.messages.last_message_index, content=token)
                        else:
                            self.messages.append('assistant', content=token, model=model)
                        yield token + self.METADATA_TOKEN + model
        
        except Exception as e:
            self.logger.error(traceback.format_exc())
            yield f'❌ _**{e.__class__.__name__}:** {e}_'

    def get_json_response(self, prompt: str, *, system_prompt: str, model: str, temperature: float) -> Generator[str, None, None]:
        framework = get_model_framework(model)
        self._prepare_for_response(prompt=prompt, system_prompt=system_prompt)

        final_message = False
        while not final_message:
            final_message = True
            response = self.clients[framework].get_json_response(model=model, temperature=temperature, messages=self.messages, tools=self.tools)
            for chunk in response:
                if isinstance(chunk, list):
                    # Only option is list of ToolCall
                    final_message = False
                    tool_calls: List[ToolCall] = chunk
                    self.messages.append(role='assistant', content=None, tool_calls=tool_calls, model=model)
                    for tool_call in tool_calls:
                        yield self.INFO_TOKEN + "Running tool: "  + tool_call.function
                        self.logger.info(f"Running tool {tool_call.function}: {str(tool_call.arguments)}", color='yellow')
                        tool_output = self.callables[tool_call.function](**tool_call.arguments)
                        self.logger.debug("Tool output:\n" + tool_output)
                        self.messages.append(role='tool', content=tool_output, tool_calls=[tool_call])
                else:
                    self.messages.append(role='assistant', content=chunk, model=model)
                    yield chunk + self.METADATA_TOKEN + model

    def export_chat(self) -> None:
        SavedChat(messages=self.messages, start_time=self.start_time, directory=self.saved_chats_dir).save_to_file()

    def load_chat(self, title: str, start_time: datetime) -> None:
        self.reset_chat()
        saved_chat = SavedChat.from_chat_title_and_time(title, start_time, directory=self.saved_chats_dir)
        self.messages = saved_chat.messages

    def get_displayed_messages(self) -> List[Message]:
        return [message for message in self.messages if message.displayed]
    
    def list_saved_chats(self) -> List[Tuple[str, datetime]]:
        chat_filenames = [f for f in os.listdir(self.saved_chats_dir) if os.path.isfile(os.path.join(self.saved_chats_dir, f)) and f.startswith(SavedChat.FILE_PREFIX) and f.endswith(SavedChat.FILE_SUFFIX)]
        saved_list = [SavedChat.filename_to_title_and_time(filename) for filename in chat_filenames]
        return sorted(saved_list, key=lambda t: t[1], reverse=True)



import sys
from argparse import ArgumentParser
from textwrap import dedent
from . import __version__
from .ui import UI
from .chatbot import Chatbot
from .settings import init_settings
from .framework_models import get_available_frameworks
from .color_logger import get_logger


def ascii_art():
    art = dedent(
        """
                                     
        ,------.                              
        |  .---' ,---.,--.  ,--.,---.  ,---.  
        |  `--, | .-. :\  `'  /| .-. :| .-. : 
        |  `---.\   --. \    / \   --.\   --. 
        `------' `----'  `--'   `----' `----'                                

        """)
    print(art)


def main():
    parser = ArgumentParser(description="Eevee Chat run-time arguments")
    parser.add_argument('-p', '--port', dest='port', default=4242, type=int, help='Port to run from', required=False)
    parser.add_argument('--version', help='Show version', dest='show_version', default=False, action='store_true')
    args = parser.parse_args()

    if args.show_version:
        print(f"Eevee Chat version: {__version__}")

    else:
        ascii_art()
        get_logger().info(f"Running version: {__version__}")
        init_settings(['../config.toml'])
        available_frameworks = get_available_frameworks()
        chatbot = Chatbot(available_frameworks)
        with UI(chatbot, available_frameworks, port=args.port):
            pass


if __name__ == '__main__':
    sys.exit(main())




import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent.resolve()

PACKAGE_NAME = "eevee"
SETUP_NAME = "eevee-chat"
AUTHOR = "Shaked Zychlinski"
AUTHOR_EMAIL = "shakedzy@gmail.com"
URL = f"http://shakedzy.xyz/{SETUP_NAME}"
DOWNLOAD_URL = f"https://pypi.org/project/{SETUP_NAME}/"

LICENSE = "CC BY-NC 4.0"
VERSION = (HERE / "VERSION").read_text(encoding="utf8").strip()
DESCRIPTION = "A single chat interface for multiple LLMs"
LONG_DESCRIPTION = (HERE / "README.md").read_text(encoding="utf8")
LONG_DESC_TYPE = "text/markdown"

requirements = (HERE / "requirements.txt").read_text(encoding="utf8")
INSTALL_REQUIRES = [s.strip() for s in requirements.split("\n")]

MIN_PYTHON_MINOR = 11
MAX_PYTHON_MINOR = 12
CLASSIFIERS = [
    f"Programming Language :: Python :: 3.{str(v)}" for v in range(MIN_PYTHON_MINOR, MAX_PYTHON_MINOR+1)
]
PYTHON_REQUIRES = f">=3.{str(MIN_PYTHON_MINOR)}"

setup(
    name=SETUP_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESC_TYPE,
    author=AUTHOR,
    license=LICENSE,
    author_email=AUTHOR_EMAIL,
    url=URL,
    download_url=DOWNLOAD_URL,
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    packages=find_packages(),
    classifiers=CLASSIFIERS,
    package_data={
        PACKAGE_NAME: ['resources/*'],  
    },
    entry_points={
        'console_scripts': [
            f'eevee = {PACKAGE_NAME}._run:main'
        ]
    }
)



import os
from typing import get_args, Set, Dict, List
from .settings import Settings
from ._types import Framework


def get_available_frameworks() -> Set[Framework]:
    available: Set[Framework] = set()
    models_dict: Dict = Settings().models
    for framework in get_args(Framework):
        if framework not in models_dict.keys():
            continue
        api_key = os.environ.get(f'{framework.upper()}_API_KEY', None)
        if api_key:
            available.add(framework)
    return available


def get_model_framework(model: str) -> Framework:
    models_dict: Dict[Framework, List[str]] = Settings().models
    for framework, models in models_dict.items():
        if model in models:
            return framework
    raise ValueError(f"Model {model} does not belong to any framework!")



def _get_version_from_setuptools() -> str:
    from pkg_resources import get_distribution

    return get_distribution("eevee-chat").version


__all__ = ["__version__"]
__version__ = _get_version_from_setuptools()



import logging
from types import TracebackType
from typing import Mapping, Dict
from ._types import Color

_DEFAULT_LOG_LEVEL = logging.INFO
_DEFAULT_LOGGER_NAME = "root"


class ColorLogger(logging.Logger):
    def __init__(self, name: str, level: int | str) -> None:
        super().__init__(name, level)
        self.propagate = False
        handler = logging.StreamHandler()
        handler.setLevel(level)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        handler.setFormatter(formatter)
        self.addHandler(handler)

    def _add_color(self, text: str, color: Color) -> str:
        reset = '\033[0m'
        colors: Dict[Color, str] = {
            'red': '\033[91m',
            'green': '\033[92m',
            'yellow': '\033[93m',
            'blue': '\033[94m',
            'magenta': '\033[95m',
            'cyan': '\033[96m'
        }
        return f'{colors[color]}{text}{reset}'

    def debug(self, msg: object, *args: object, color: Color | None = None, exc_info: bool | tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None] | BaseException | None = None, stack_info: bool = False, stacklevel: int = 1, extra: Mapping[str, object] | None = None) -> None:
        if color:
            msg = self._add_color(str(msg), color)
        return super().debug(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)
    
    def info(self, msg: object, *args: object, color: Color | None = None, exc_info: bool | tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None] | BaseException | None = None, stack_info: bool = False, stacklevel: int = 1, extra: Mapping[str, object] | None = None) -> None:
        if color:
            msg = self._add_color(str(msg), color)
        return super().info(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)
    
    def warn(self, msg: object, *args: object, color: Color | None = None, exc_info: bool | tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None] | BaseException | None = None, stack_info: bool = False, stacklevel: int = 1, extra: Mapping[str, object] | None = None) -> None:
        if color:
            msg = self._add_color(str(msg), color)
        return super().warn(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)
 
    def warning(self, msg: object, *args: object, color: Color | None = None, exc_info: bool | tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None] | BaseException | None = None, stack_info: bool = False, stacklevel: int = 1, extra: Mapping[str, object] | None = None) -> None:
        if color:
            msg = self._add_color(str(msg), color)
        return super().warning(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)

    def error(self, msg: object, *args: object, color: Color | None = None, exc_info: bool | tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None] | BaseException | None = None, stack_info: bool = False, stacklevel: int = 1, extra: Mapping[str, object] | None = None) -> None:
        if color:
            msg = self._add_color(str(msg), color)
        return super().error(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)
    
    def critical(self, msg: object, *args: object, color: Color | None = None, exc_info: bool | tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None] | BaseException | None = None, stack_info: bool = False, stacklevel: int = 1, extra: Mapping[str, object] | None = None) -> None:
        if color:
            msg = self._add_color(str(msg), color)
        return super().critical(msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extra=extra)
    

def get_logger(name: str| None = None, level: str | int | None = None):
    return ColorLogger(name or _DEFAULT_LOGGER_NAME, level=level or _DEFAULT_LOG_LEVEL)


from dynaconf import Dynaconf
from typing import List
from .color_logger import get_logger


_CONFIG_FILES_PATHS: List[str] = []


def init_settings(config_files: List[str]) -> None:
    global _CONFIG_FILES_PATHS
    _CONFIG_FILES_PATHS = config_files


class Settings:
    """
    A Singleton class for settings
    """
    _instance = None
    _settings = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            if not _CONFIG_FILES_PATHS:
                raise ValueError('No config files supplied to an uninitialized instance!')
            else:
                get_logger().info(f'Loading config files: {str(_CONFIG_FILES_PATHS)}')
                cls._settings = Dynaconf(
                    settings_files=_CONFIG_FILES_PATHS,
                    environments=False,
                    merge_enabled=True
                )
            cls._instance = super(Settings, cls).__new__(cls)
        return cls._instance

    def __getattr__(self, name):
        try:
            return getattr(self._settings, name)
        except AttributeError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")
    
    def __setattr__(self, name, value):
        if name == '_settings':
            # Handle setting the '_settings' attribute directly.
            # Use super().__setattr__() to avoid infinite recursion.
            super().__setattr__(name, value)
        else:
            # Attempt to set the attribute on the '_settings' object.
            try:
                setattr(self._settings, name, value)
            except AttributeError:
                # If it fails, set the attribute normally.
                super().__setattr__(name, value)
